#include "EntityGenerator.h"

#include "EntityType.h"
#include "Game.h"
#include "JNoise.h"
#include "JsonExpression.h"

EntityGenerator::EntityGenerator()
    : ReferenceCounter(),
      noise(0),
      noiseConfig(0),
      threshold(0.0),
      zType(0),
      condition(0)
{}

EntityGenerator::~EntityGenerator()
{
    if (condition)
    {
        condition->release();
    }
    if (noise)
    {
        noise->release();
    }
    if (noiseConfig)
    {
        noiseConfig->release();
    }
}

void EntityGenerator::initialize(JExpressionMemory* zMemory)
{
    if (noiseConfig)
    {
        if (noise) noise->release();
        noise = JNoise::parseNoise(noiseConfig, zMemory);
    }
}

bool EntityGenerator::isGenerated(
    int x, int y, int z, int dimensionId, JExpressionMemory* zMemory)
{
    return (!noise
               || noise->getNoise((double)x, (double)y, (double)z) <= threshold)
        && condition->getValue(zMemory);
}

Entity* EntityGenerator::generate(Framework::Vec3<float> pos, int dimesnionId)
{
    return zType->createEntityAt(pos, dimesnionId);
}

EntityGeneratorFactory::EntityGeneratorFactory()
    : ObjectTypeFactory()
{}

EntityGenerator* EntityGeneratorFactory::fromJson(
    Framework::JSON::JSONObject* zJson) const
{
    EntityGenerator* result = new EntityGenerator();
    if (zJson->hasValue("noise"))
    {
        result->noiseConfig = zJson->getValue("noise")->asObject();
    }
    if (zJson->hasValue("threshold"))
    {
        result->threshold = zJson->zValue("threshold")->asNumber()->getNumber();
    }
    if (zJson->hasValue("type"))
    {
        result->zType
            = Game::INSTANCE->zEntityType(Game::INSTANCE->getEntityTypeId(
                zJson->zValue("type")->asString()->getString()));
    }
    if (zJson->hasValue("condition"))
    {
        result->condition
            = Game::INSTANCE->zTypeRegistry()->fromJson<JBoolExpression>(
                zJson->zValue("condition"));
    }
    return result;
}

Framework::JSON::JSONObject* EntityGeneratorFactory::toJsonObject(
    EntityGenerator* zObject) const
{
    Framework::JSON::JSONObject* result = new Framework::JSON::JSONObject();
    if (zObject->noiseConfig)
    {
        result->addValue("noise",
            dynamic_cast<Framework::JSON::JSONValue*>(
                zObject->noiseConfig->getThis()));
    }
    result->addValue(
        "threshold", new Framework::JSON::JSONNumber(zObject->threshold));
    result->addValue(
        "type", new Framework::JSON::JSONString(zObject->zType->getName()));
    result->addValue("condition",
        Game::INSTANCE->zTypeRegistry()->toJson(zObject->condition));
    return result;
}

JSONObjectValidationBuilder* EntityGeneratorFactory::addToValidator(
    JSONObjectValidationBuilder* builder) const
{
    Framework::RCArray<Framework::Text> entityTypeNames;
    for (int i = 0; i < Game::INSTANCE->getEntityTypeCount(); i++)
    {
        if (Game::INSTANCE->zEntityType(i))
        {
            entityTypeNames.add(
                new Framework::Text(Game::INSTANCE->zEntityType(i)->getName()));
        }
    }
    // TODO: add EntityTypeNameFactory
    return builder->withRequiredAttribute("noise", JNoise::getValidator(true))
        ->withRequiredNumber("threshold")
        ->whichIsOptional()
        ->whichIsGreaterOrEqual(0.0)
        ->whichIsLessOrEqual(1.0)
        ->finishNumber()
        ->withRequiredString("type")
        ->whichIsOneOf(entityTypeNames)
        ->finishString()
        ->withRequiredAttribute("condition",
            Game::INSTANCE->zTypeRegistry()->getValidator<JBoolExpression>());
}
